/*
 * AES XTS encryption routines in userspace
 *
 * Copyright (C) 2013 Johannes Goetzfried <johannes@jgoetzfried.de>
 *
 * based on 'storage/lib/crypto/aes_xts.c' from bitvisor 1.3 (see below)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 *
 *
 * Copyright (c) 2007, 2008 University of Tsukuba
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Tsukuba nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "aes_xts.h"


static void inline xor128(void *dst, const void *src1, const void *src2)
{
	const uint64_t *s1 = src1;
	const uint64_t *s2 = src2;
	uint64_t *d = dst;

	d[0] = s1[0] ^ s2[0];
	d[1] = s1[1] ^ s2[1];
}

static void inline movzx128(void *dst, const uint64_t src)
{
	uint64_t *d = dst;

	d[0] = src;
	d[1] = 0;
}

static int shl128(void *dst, const void *src)
{
	int c0, c1;
	const uint64_t *s = src;
	uint64_t *d = dst;

	c0 = s[0] & (1ULL << 63) ? 1 : 0;
	c1 = s[1] & (1ULL << 63) ? 1 : 0;
	d[0] = s[0] << 1;
	d[1] = s[1] << 1 | c0;
	return c1;
}

static void gf_mul128(uint8_t *dst, const uint8_t *src)
{
	int carry;
	static const uint8_t gf_128_fdbk = 0x87;

	carry = shl128(dst, src);
	if (carry)
		dst[0] ^= gf_128_fdbk;
}

static void aes_xts_crypt(uint8_t *dst, const uint8_t *src, aes_crypt_func_t crypt,
			  const AES_KEY *tk, const AES_KEY *ck, uint64_t lba, uint32_t sector_size)
{
	int i;
	uint8_t tweak[AES_BLK_BYTES];

	if (sector_size % AES_BLK_BYTES != 0) {
		fprintf(stderr, "FATAL: Sector size is no multiple of block size!\n");
		exit(EXIT_FAILURE);
	}

	movzx128(tweak, lba);				// convert sector number to tweak plaintext
	AES_encrypt(tweak, tweak, tk);			// encrypt the tweak
	for (i = sector_size; i > 0; i -= AES_BLK_BYTES) {
		xor128(dst, src, tweak);		// merge the tweak into the input block
		crypt(dst, dst, ck);			// encrypt one block
		xor128(dst, dst, tweak);
		gf_mul128(tweak, tweak);
		dst += AES_BLK_BYTES;
		src += AES_BLK_BYTES;
	}
}

void aes_xts_encrypt(uint8_t *dst, const uint8_t *src,
		     const struct aes_xts_keyctx *k, uint64_t lba, uint32_t sector_size)
{
	aes_xts_crypt(dst, src, AES_encrypt, &k->encrypt.tweak_key,
		      &k->encrypt.encrypt_key, lba, sector_size);
}

void aes_xts_decrypt(uint8_t *dst, const uint8_t *src,
		     const struct aes_xts_keyctx *k, uint64_t lba, uint32_t sector_size)
{
	aes_xts_crypt(dst, src, AES_decrypt, &k->decrypt.tweak_key,
		      &k->decrypt.decrypt_key, lba, sector_size);
}

struct aes_xts_keyctx *aes_xts_setkey(const uint8_t *key, int bits)
{
	int keybit = bits / 2;
	int keylen = keybit / 8;
	struct aes_xts_keyctx *keyctx = malloc(sizeof(struct aes_xts_keyctx));

	if (keyctx == NULL) {
		fprintf(stderr, "FATAL: Memory allocation failed!\n");
		exit(EXIT_FAILURE);
	}

	AES_set_encrypt_key(key + keylen, keybit, &keyctx->encrypt.tweak_key);
	AES_set_encrypt_key(key         , keybit, &keyctx->encrypt.encrypt_key);
	AES_set_encrypt_key(key + keylen, keybit, &keyctx->decrypt.tweak_key);
	AES_set_decrypt_key(key         , keybit, &keyctx->decrypt.decrypt_key);

	return keyctx;
}

void free_aes_xts_keyctx(struct aes_xts_keyctx *keyctx)
{
	free(keyctx);
}
