# MARK - User Interface Dialogs
#
# Copyright (C) 2013  Johannes Goetzfried <johannes@jgoetzfried.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time;
import curses;
import curses.panel;
from basic import *

ACTION_OK     = 1;
ACTION_REPEAT = 2;
ACTION_EXIT   = 3;

COLOR_ACTIVE  = curses.COLOR_YELLOW;
COLOR_DONE    = curses.COLOR_GREEN;
COLOR_FAIL    = curses.COLOR_RED;


class Header(Panel):
	def __init__(self, title):
		self.hl = title;

		sh, sw = ui().gethw();
		Panel.__init__(self, 3, sw, 0, 0);

	def draw(self):
		sh, sw = ui().gethw();
		self.background(curses.COLOR_WHITE, curses.COLOR_RED);
		self.writeto(1, max(1, (sw - len(self.hl)) / 2), self.hl);
		ui().update();


class MainPanel(CenterPanel):
	def __init__(self, title, actions, w=0):
		self.actions     = actions;
		self.key_length  = max(map(lambda x: len(x[0]), actions));
		self.desc_length = max(map(lambda x: len(x[2]), actions));

		height = 2 + 2 * len(actions) + 1;
		width = max(6 + self.key_length + 2 + self.desc_length + 6, w);
		CenterPanel.__init__(self, height, width, title);

	def draw(self):
		self.background(curses.COLOR_YELLOW, curses.COLOR_BLUE);

		for a in self.actions:
			self.writeto(2 + 2 * self.actions.index(a), 6, ' ' * (self.key_length - len(a[0])) + a[0]);
			self.writeto(2 + 2 * self.actions.index(a), 8 + self.key_length, a[2]);

		ui().update();

		while True:
			ch = ui().getch();
			for a in self.actions:
				if ch == a[1]:
					a[3]();

			ui().handlekey(ch);


class StepPanel(CenterPanel):
	def __init__(self, title, actions, w=0):
		self.actions      = actions;
		self.current_step = 0;
		self.line_length  = 0;
		self.strlist      = [];

		dlist = filter(lambda x: len(x[0]) > 0, actions);
		self.strlist = map(lambda x: x[0], dlist);
		height = 2 + 2 * len(self.strlist) + 1;
		self.line_length = max(map(len, self.strlist));
		width = max(3 + self.line_length + 2 + 20 + 3, w);
		CenterPanel.__init__(self, height, width, title);

	def draw(self):
		self.background(curses.COLOR_WHITE, curses.COLOR_BLACK);

		for s in self.strlist:
			self.writeto(2 + 2 * self.strlist.index(s), 3, ' ' * (self.line_length - len(s)) + s + ':');

		ui().update();

		for a in self.actions:
			result = ACTION_REPEAT;
			while result == ACTION_REPEAT:
				if len(a[0]) > 0:
					self.status(COLOR_ACTIVE, "processing");
					time.sleep(0.2);
				result = a[1](self);

			if result == ACTION_OK:
				if len(a[0]) > 0:
					self.status(COLOR_DONE, "done"); 
			else:
				self.hide();
				ui().update();
				return;

			if len(a[0]) > 0:
				self.current_step = self.current_step + 1;

		self.hide();
		ui().update();

	def status(self, fg, msg):
		h, w = self.gethw();
		self.writeto(2 + 2 * self.current_step, 5 + self.line_length, ' ' * (w - 8 - self.line_length));
		self.fg(fg);
		self.writeto(2 + 2 * self.current_step, 5 + self.line_length, msg);
		ui().update();


def textdim(text):
	if len(text) < 1:
		return 0, 0;

	l = text.split('\n');
	h = len(l);
	w = max(map(len, l));

	return h, w;


def choicebox(title, prompt, choices):
	if len(choices) < 1:
		return -1;

	h, w = textdim(prompt);
	if h > 0:
		h = h + 1

	# padding, prompt, choices, padding
	height = 1 + h + len(choices) + 1;

	width = max(4 + len(title), w);
	for c in choices:
		width = max(width, 4 + len(c));

	panel = CenterPanel(height + 2, width + 4, title);
	panel.background(curses.COLOR_YELLOW, curses.COLOR_BLUE);

	panel.writeto(2, 2, prompt);

	choice = 0;
	while True:
		for c in choices:
			if choices.index(c) == choice:
				l = '> ' + c + ' <';
			else:
				l = '  ' + c + '  ';

			panel.writeto(2 + h + choices.index(c), (4 + width - len(l)) / 2, l);
			ui().update();

		ch = ui().getch();

		if ch == 10:
			break;

		if ch == curses.KEY_UP:
			choice = (choice - 1 + len(choices)) % len(choices);
		elif ch == curses.KEY_DOWN:
			choice = (choice + 1) % len(choices);
		else:
			ui().handlekey(ch);

	panel.hide();
	ui().update();
	return choice;


def inputbox(title, prompt, h, w, password=False, skip=False, prefill=''):
	def get_unicode_char(c):
		def get_check_next_byte():
			c = ui().getch();
			if 128 <= c <= 191:
				return c;
			else:
				raise UnicodeError;

		b = []
		if c <= 127:
			b.append(c);
		elif 194 <= c <= 223:
			b.append(c);
			b.append(get_check_next_byte());
		elif 224 <= c <= 239:
			b.append(c);
			b.append(get_check_next_byte());
			b.append(get_check_next_byte());
		elif 240 <= c <= 244:
			b.append(c);
			b.append(get_check_next_byte());
			b.append(get_check_next_byte());
			b.append(get_check_next_byte());
		buf = ''.join([chr(x) for x in b])
		return buf.decode('utf-8')

	if skip == True:
		prompt = prompt + '\n(Or press F10 to skip this step.)';

	prompth, promptw = textdim(prompt);
	prompth = prompth + 1;

	height = 1 + prompth + h + 1;
	width = max(6 + len(title), promptw + 2);
	width = max(width, w)

	bpanel = CenterPanel(height + 2, width + 2, title);
	bpanel.background(curses.COLOR_YELLOW, curses.COLOR_BLUE);

	bpanel.writeto(2, 2, prompt);
	bpanel.go(2 + prompth, 1);
	sy, sx = bpanel.getscreenyx();

	tpanel = Panel(h, w, sy, sx, False);
	tpanel._win.keypad(1);
	tpanel.background(curses.COLOR_BLACK, curses.COLOR_WHITE);
	curses.curs_set(1);
	ui().update();

	result = u'';
	ch = ui().getch();
	while ch != 10:
		if ch == curses.KEY_BACKSPACE:
			result = result[:-1];
		elif ch == curses.KEY_F10 and skip == True:
			result = u'';
			break;
		elif ch >= curses.KEY_MIN:
			ui().handlekey(ch);
		else:
			result = result + get_unicode_char(ch);

		tpanel.erase();
		if password == True:
			tpanel.writeto(0, 0, '*' * len(result), h);
		else:
			tpanel.writeto(0, 0, result, h);

		ui().update();
		ch = ui().getch();

	curses.curs_set(0);
	tpanel.hide();
	bpanel.hide();
	ui().update();

	return result;


def retry(msg):
	if choicebox('Retry?', msg, ['Try again', 'Return to main menu']) == 0:
		return ACTION_REPEAT;
	else:
		return ACTION_EXIT;

